/*
 * Copyright (C) 2011-2013 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.bootleg;

/**
 * Bootleg framework configuration.
 * You can specify your custom configuration by declaring {@code ServletContext} 
 * init parameter in web.xml as follow:
 * <pre>
 * &lt;context-param&gt;
 *     &lt;param-name&gt;org.eiichiro.bootleg.configuration&lt;/param-name&gt;
 *     &lt;param-value&gt;com.example.your.app.CustomConfiguration&lt;/param-value&gt;
 * &lt;/context-param&gt;
 * </pre>
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public interface Configuration {

	/**
	 * Specifies URL routing configuration of the current application.
	 * 
	 * @param routing URL routing configuration of the current application.
	 * @see Routing
	 */
	public void routing(Routing routing);
	
	/**
	 * Specifies HTTP request processing pipeline configuration of the current 
	 * application.
	 * 
	 * @param pipeline HTTP request processing pipeline container.
	 */
	public void pipeline(Pipeline<WebContext> pipeline);
	
	/**
	 * Returns {@link Request} implementation class to represent the specified 
	 * content type.
	 * 
	 * @param contentType The content type specified in HTTP request header. 
	 * This parameter may be <code>null</code>.
	 * @return The {@code Request} implementation class corresponding to the 
	 * specified content type.
	 */
	public Class<? extends Request> requestType(String contentType);
	
	/**
	 * Returns {@link Response} implementation class to represent the specified 
	 * content type.
	 * 
	 * @param contentType The content type specified in HTTP response header. 
	 * This parameter may be <code>null</code>.
	 * @return The {@code Response} implementation class corresponding to the 
	 * specified content type.
	 */
	public Class<? extends Response> responseType(String contentType);
	
	/**
	 * Returns {@link Loader} instance to load Web endpoint classes.
	 * 
	 * @return The loader instance to load Web endpoint classes.
	 */
	public Loader loader();
	
	/**
	 * Returns {@link Instantiator} instance to instantiate Web endpoint 
	 * instance.
	 * 
	 * @return The instantiator instance to instantiate Web endpoint instance.
	 */
	public Instantiator instantiator();
	
}
