/*
 * Copyright (C) 2011-2013 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.bootleg;

import java.util.HashMap;
import java.util.Map;

import org.eiichiro.bootleg.json.JSONRequest;
import org.eiichiro.bootleg.json.JSONResponse;
import org.eiichiro.bootleg.xml.XMLRequest;
import org.eiichiro.bootleg.xml.XMLResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Bootleg framework's default configuration.
 * If you would not specify your own configuration in web.xml, Bootleg uses this 
 * configuration by default.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class DefaultConfiguration implements Configuration {

	private Logger logger = LoggerFactory.getLogger(getClass());
	
	private Map<String, Class<? extends Request>> requestTypes = new HashMap<String, Class<? extends Request>>();
	
	private Map<String, Class<? extends Response>> responseTypes = new HashMap<String, Class<? extends Response>>();
	
	private Loader loader = new ClasspathLoader();
	
	private Instantiator instantiator = new Instantiator() {

		public <T> T instantiate(Class<T> endpoint) {
			try {
				return endpoint.newInstance();
			} catch (Exception e) {
				logger.warn("Failed to instantiate Web endpoint", e);
			}
			
			return null;
		}
		
	};
	
	/**
	 * Constructs a new {@code DefaultConfiguration} instance.
	 * Sets up 'Content-type'-{@link Request} type mapping and 
	 * 'Content-type'-{@link Response} type mapping.
	 */
	public DefaultConfiguration() {
		requestTypes.put(MediaType.APPLICATION_JSON, JSONRequest.class);
		requestTypes.put(MediaType.APPLICATION_XML, XMLRequest.class);
		requestTypes.put(MediaType.TEXT_HTML, GenericRequest.class);
		requestTypes.put(MediaType.TEXT_PLAIN, GenericRequest.class);
		responseTypes.put(MediaType.APPLICATION_JSON, JSONResponse.class);
		responseTypes.put(MediaType.APPLICATION_XML, XMLResponse.class);
		responseTypes.put(MediaType.TEXT_HTML, GenericResponse.class);
		responseTypes.put(MediaType.TEXT_PLAIN, GenericResponse.class);
	}
	
	/**
	 * Specifies the default ignore-routing patterns.
	 * Bootleg ignores static resources (<code>*.js</code>, <code>*.css</code>, 
	 * <code>*.gif</code>, <code>*.png</code>, <code>*.jpg</code>, 
	 * <code>*.jsp</code>, <code>*.html</code>) by default.
	 * 
	 * @param routing {@link Routing} configuration of the current application.
	 */
	public void routing(Routing routing) {
		routing.ignore("*.js", "*.css", "*.gif", "*.png", "*.jpg", "*.jsp", "*.html");
	}
	
	/**
	 * Specifies the default HTTP request processing pipeline configuration.
	 * HTTP request processing pipeline consists of [Route] -&gt; [Receive] 
	 * -&gt; [Invoke] -&gt; [Send] by default.
	 * 
	 * @param pipeline HTTP request processing pipeline container.
	 * @see Route
	 * @see Receive
	 * @see Invoke
	 * @see Send
	 */
	public void pipeline(Pipeline<WebContext> pipeline) {
		pipeline.stage(Route.class.getSimpleName(), new Route())
				.stage(Receive.class.getSimpleName(), new Receive())
				.stage(Invoke.class.getSimpleName(), new Invoke())
				.stage(Send.class.getSimpleName(), new Send());
	}

	/**
	 * Returns the default request type corresponding to the specified MIME 
	 * media type.
	 * 'Content-type'-{@link Request} type mapping is: 
	 * <ul>
	 * <li><code>application/xml</code> -&gt; {@code XMLRequest.class}</li>
	 * <li><code>application/json</code> -&gt; {@code JSONRequest.class}</li>
	 * <li><code>text/plain</code> -&gt; {@code GenericRequest.class}</li>
	 * <li><code>text/html</code> -&gt; {@code GenericRequest.class}</li>
	 * <li>The others (including <code>null</code>) -&gt; {@code GenericRequest.class}</li>
	 * </ul>
	 * 
	 * @param contentType The content type specified in HTTP request header. 
	 * This parameter may be <code>null</code>.
	 * @return The {@link Request} implementation class corresponding to the 
	 * specified content type.
	 */
	public Class<? extends Request> requestType(String contentType) {
		Class<? extends Request> requestType = requestTypes.get(contentType);
		return (requestType != null) ? requestType : GenericRequest.class;
	}

	/**
	 * Returns the default response type corresponding to the specified MIME 
	 * media type.
	 * 'Content-type'-{@link Response} type mapping is: 
	 * <ul>
	 * <li><code>application/xml</code> -&gt; {@code XMLResponse.class}</li>
	 * <li><code>application/json</code> -&gt; {@code JSONResponse.class}</li>
	 * <li><code>text/plain</code> -&gt; {@code GenericResponse.class}</li>
	 * <li><code>text/html</code> -&gt; {@code GenericResponse.class}</li>
	 * <li>The others (including <code>null</code>) -&gt; {@code GenericResponse.class}</li>
	 * </ul>
	 * 
	 * @param contentType The content type specified in HTTP response header. 
	 * This parameter may be <code>null</code>.
	 * @return The {@code Response} implementation class corresponding to the 
	 * specified content type.
	 */
	public Class<? extends Response> responseType(String contentType) {
		Class<? extends Response> responseType = responseTypes.get(contentType);
		return (responseType != null) ? responseType : GenericResponse.class;
	}

	/**
	 * Returns cached {@link ClasspathLoader} instance.
	 * 
	 * @return The cached {@code ClasspathLoader} instance.
	 */
	public Loader loader() {
		return loader;
	}

	/**
	 * Returns cached {@link Instantiator} instance that instantiates Web 
	 * endpoint class by {@code Class#newInstance()}.
	 * 
	 * @return The cached {@code Instantiator} instance.
	 */
	public Instantiator instantiator() {
		return instantiator;
	}

}
